from abc import ABC, abstractmethod


class BaseServer(ABC):
    def __init__(
        self,
        rmq_urls: list[str],
        service_name: str,
        queue_error_name: str = "error",
    ):
        self.rmq_urls = rmq_urls
        self.service_name = service_name
        self.queue_error_name = queue_error_name

    @classmethod
    @abstractmethod
    async def create(
        cls,
        rmq_urls: list[str],
        service_name: str,
        queue_error_name: str = "error",
        **kwargs,
    ):
        """
        The method should be a factory for asynchronous
        initialization of the class
        """
        raise NotImplementedError

    @abstractmethod
    async def bind(self, *args, **kwargs):
        """
        The method should bind
        methods to abstract routing
        """
        raise NotImplementedError

    @abstractmethod
    async def shutdown(self):
        raise NotImplementedError



import json
import logging
import uuid
from datetime import datetime
from typing import Literal, Optional, Union

import pytz
from aio_pika import IncomingMessage
from pydantic import BaseModel, ValidationError, create_model

from aioka.server import BaseMeta, MessageBase, RequestMeta, ResponseMeta
from aioka.server.errors.fmt import gen_error_message

from .base import BaseAction

logger = logging.getLogger(__name__)


class Action(BaseAction):
    REQUEST_META = RequestMeta
    REQUEST_PAYLOAD: BaseModel

    RESPONSE_META = ResponseMeta
    RESPONSE_PAYLOAD: BaseModel

    def __init__(self, action_name: str):
        super(Action, self).__init__(action_name=action_name)

        self.REQUEST_META = create_model(
            "REQUEST_META",
            __base__=self.REQUEST_META,
            action=(Literal[self.action_name], ...),
        )
        self.REQUEST_META_VALIDATION = create_model(
            "REQUEST_META_VALIDATION",
            meta=(self.REQUEST_META, ...),
            payload=(Optional[dict], None),
        )

    @property
    def queue_name(self):
        return fmt_queue_name(
            self.action_name, service_name=self._service_name
        )

    @property
    def routing_keys(self):
        return [self.queue_name]

    async def consume(self, message: IncomingMessage):
        async with message.process(
            requeue=True,
            reject_on_redelivered=False,
            ignore_processed=True,
        ):
            # 1) проверяем валидность схемы
            # 1.1) проверяем что пришел json
            try:
                json_data = json.loads(message.body)
            except json.JSONDecodeError as e:
                logger.error("Is not JSON", exc_info=True)
                error_message = gen_error_message(message.body, e)
                await self.publish(message=error_message)
                return
            # 1.2 десериализовать сообщение
            # 1.3 проверить корректность меты c точки зрения общей схемы
            try:
                message_obj = self.REQUEST_META_VALIDATION.model_validate(
                    json_data
                )
                meta_obj = message_obj.meta
                assert isinstance(meta_obj, self.REQUEST_META)

            except ValidationError as e:
                logger.error("Meta is not valid", exc_info=True)
                error_message = gen_error_message(message.body, e)
                await self.publish(message=error_message)
                return
            # 1.4 проверить корректность payload
            try:
                payload_obj = self.REQUEST_PAYLOAD.parse_obj(
                    message_obj.payload
                )
            except ValidationError as e:
                logger.error("Payload is not valid", exc_info=True)
                error_message = gen_error_message(message.body, e)
                await self.publish(message=error_message)
                if meta_obj.reply_to:
                    await self.publish(
                        routing_key=meta_obj.reply_to, message=error_message
                    )
                return
            # Выполнить
            try:
                RESPONSE_TYPE = self.RESPONSE_PAYLOAD
                response: dict | RESPONSE_TYPE | None = await self(
                    meta=meta_obj, payload=payload_obj
                )
                if meta_obj.reply_to and response is not None:
                    await self.publish(
                        routing_key=meta_obj.reply_to,
                        message=RESPONSE_TYPE.model_validate(response),
                    )
            except Exception as e:
                logger.error(
                    f"Unhandled exception in {self.__call__}", exc_info=True
                )
                error_message = gen_error_message(
                    message.body, e, error_type="internal"
                )
                await self.publish(message=error_message)
                if meta_obj.reply_to:
                    await self.publish(
                        routing_key=meta_obj.reply_to, message=error_message
                    )

    async def validate_message(
        self, body: bytes
    ) -> Union[bool, tuple[BaseModel, BaseModel]]:
        ...

    async def validate_meta(self, body: bytes):
        ...

    async def validate(self, body: bytes):
        ...

    async def consume_success(
        self, meta: BaseMeta, response_payload: BaseModel
    ):
        response_meta = self.RESPONSE_META.model_validate(
            {
                "from": self._service_name,
                "reply_to": None,
                "service": meta.service,
                "action": meta.action,
                "sent_at": datetime.now(tz=pytz.utc),
                "message_uuid": uuid.uuid4(),
                "reply_to_message_uuid": meta.message_uuid,
                "path": meta.path + [self._service_name],
            }
        )
        message = MessageBase(meta=response_meta, payload=response_payload)
        logger.debug(
            f"consume success reply_to: {meta.reply_to}, message: {message}"
        )
        await self.publish(routing_key=meta.reply_to, message=message)

    async def publish(
        self,
        message: Union[BaseModel, bytes],
        routing_key: Optional[str] = None,
    ):
        if isinstance(message, BaseModel):
            message_body = message.model_dump_json(
                indent=2, by_alias=True
            ).encode()
        elif isinstance(message, str):
            message_body = message.encode()
        else:
            message_body = message
        await self.client(
            message=message_body,
            routing_key=routing_key,
        )

    async def __call__(self, meta: BaseMeta, payload: BaseModel):
        raise NotImplementedError


def fmt_queue_name(handle_name: str, service_name: str):
    return f"{service_name}.{handle_name}"



import asyncio
import logging
import threading
import time
from contextlib import asynccontextmanager, contextmanager
from typing import Optional, Type

import uvicorn
from aio_pika import Connection
from uvicorn._types import ASGIApplication

from aioka.action.base import BaseAction
from aioka.handler.base import BaseHandler
from aioka.server.base import BaseServer

logger = logging.getLogger(__name__)


class UvicornServer(uvicorn.Server):
    def install_signal_handlers(self):
        pass

    @contextmanager
    def run_in_thread(self):
        thread = threading.Thread(target=self.run)
        thread.start()
        try:
            while not self.started:
                time.sleep(1e-3)
            yield
        finally:
            self.should_exit = True
            thread.join()


class AsyncApp:
    def __init__(self, server: BaseServer):
        self.server = server
        self._handlers: list[BaseHandler] = []
        self.default_handler = BaseHandler()
        self._loop = asyncio.get_event_loop()

        self.uvicorn_asgi: Optional[uvicorn.Server] = None
        self.rq_conn: Optional[Connection] = None

    def run(self):
        logger.info("%s try start", self.server.service_name)
        try:
            self._loop.run_forever()
        except KeyboardInterrupt:
            pass
        finally:
            logger.info("%s stopped", self.server.service_name)

    async def bind(self):
        logger.debug(self._handlers)

        all_handlers = [self.default_handler] + self._handlers

        for handler in all_handlers:
            logging.info("binding handler: %s" % handler)
            handler = await handler.create(self.server)
            await handler.bind()

    def add_handler(self, handler: BaseHandler):
        self._handlers.append(handler)

    def add_action(self, name: str, action: BaseAction):
        self.default_handler.add_action(action_name=name, action=action)

    def add_uvicorn_asgi(self, app: Type[ASGIApplication], **kwargs):
        config = uvicorn.Config(app, lifespan="on", **kwargs)
        server = UvicornServer(config=config)
        self._uvicorn_asgi_flag = server.run_in_thread()
        self._loop.create_task(server.serve())
        self.uvicorn_asgi = server

    async def shutdown(self):
        """
        try to shut down gracefully
        """
        if self.uvicorn_asgi:  # type: uvicorn.Server
            await self.uvicorn_asgi.shutdown()
            next(self._uvicorn_asgi_flag)
        await self.server.shutdown()

        logger.info("Stopping server")


@asynccontextmanager
async def create_app(server: Type[BaseServer], **kwargs) -> AsyncApp:
    server = await server.create(**kwargs)
    app = AsyncApp(server=server)
    await app.bind()
    return app



import json
import logging
import traceback
import uuid
from datetime import datetime
from json.encoder import JSONEncoder
from typing import Any, Literal

import pytz
from pydantic import BaseModel

from aioka.server.errors.schema import ExceptionSchema
from aioka.server.schema import BaseMeta

logger = logging.getLogger(__name__)


def create_message(meta: dict, payload: dict):
    return {
        "meta": meta,
        "payload": payload,
    }


def create_error_response_meta(service_name: str):
    return {
        "from": service_name,
        "reply_to": None,
        "sent_at": datetime.now(tz=pytz.utc),
        "message_uuid": uuid.uuid4(),
        "reply_to_message_uuid": None,
    }


class JsonEncoder(JSONEncoder):
    # fixme мне кажется где-то я этот клас уже видел в таком исполнении
    def default(self, o: Any) -> Any:
        if isinstance(o, datetime):
            return o.isoformat()
        elif isinstance(o, uuid.UUID):
            return str(o)
        return super().default(o)


def gen_error_message(
    raw_message,
    e,
    error_type: Literal["internal", "validation"] = "validation",
):
    if error_type == "validation":
        if hasattr(e, "errors") and callable(e.errors):
            details = e.errors()
        else:
            details = []

        error = generate_error(
            error_type=error_type,
            validation_details={
                "error": str(e),
                "trace": traceback.format_exc(),
                "details": details,
            },
            raw_message=raw_message,
        )
    else:
        error = generate_error(
            error_type=error_type,
            internal_details={
                "error": str(e),
                "trace": traceback.format_exc(),
            },
            raw_message=raw_message,
        )

    meta_response = create_error_response_meta()

    error_message = json.dumps(
        create_message(meta_response, error.model_dump(by_alias=True)),
        indent=2,
        ensure_ascii=False,
        cls=JsonEncoder,
    ).encode()

    return error_message


def generate_meta(meta: BaseModel | dict, service_name: str) -> BaseMeta:
    if isinstance(meta, BaseModel):
        meta_dict: dict = meta.model_dump(by_alias=True)
    else:
        meta_dict = meta
    return BaseMeta.model_validate(
        {
            **meta_dict,
            "from": service_name,
            "reply_to": None,
            "sent_at": datetime.now(tz=pytz.utc),
            "message_uuid": str(uuid.uuid4()),
            "reply_to_message_uuid": meta_dict.get("message_uuid"),
        }
    )


def generate_error(
    raw_message: str = None,
    response_ok: bool = True,
    error_type: Literal["internal", "validation"] = "internal",
    validation_details=None,
    internal_details=None,
) -> ExceptionSchema:
    if validation_details is None:
        validation_details = {}
    if internal_details is None:
        internal_details = {}

    return ExceptionSchema.model_validate(
        {
            "response_ok": response_ok,
            "error_type": error_type,
            "validation_details": validation_details,
            "internal_details": internal_details,
            "raw_message": raw_message,
        }
    )



from datetime import datetime
from uuid import UUID, uuid4

from pydantic import BaseModel, Field


class BaseMeta(BaseModel):
    source: str = Field(description="service name (queue)")
    reply_to: str | None = Field(
        description="reply to queue name, if no reply None"
    )
    service: str = Field(description="another service name")
    action: str = Field(..., description="another service action name")
    sent_at: datetime = Field(
        default_factory=datetime.utcnow,
        description="sent at (datetime with tz)",
    )
    message_uuid: UUID = Field(
        default_factory=uuid4, description="uuid4 of this message"
    )
    reply_to_message_uuid: UUID | None = Field(
        None,
        description="uuid4 reply message, if first message then None",
    )
    path: list[str] = Field(
        default_factory=list, description="message reply history"
    )


class RequestMeta(BaseMeta):
    reply_to: str
    service: str
    reply_to_message_uuid: UUID | None = None


class ResponseMeta(BaseMeta):
    reply_to: str | None = None


class RequestMessageBase(BaseModel):
    meta: RequestMeta
    payload: dict | None = None


class MessageBase(BaseModel):
    meta: BaseMeta
    payload: BaseModel



from abc import ABC, abstractmethod
from typing import Callable, Coroutine, Optional

from aio_pika import IncomingMessage


class BaseAction(ABC):
    """
    Base class for actions called by routes

    :param action_name: Procedure name to indicate in routing
    """

    def __init__(self, action_name: str):
        self.action_name = action_name
        self.client: Optional[Callable[..., Coroutine[any, any, any]]] = None
        self._service_name: str | None = None

    @property
    def service_name(self):
        return self._service_name

    @service_name.setter
    def service_name(self, value: str):
        if isinstance(value, str):
            self._service_name = value
        raise ValueError("service_name must be str")

    @abstractmethod
    async def __call__(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    async def consume(self, message: IncomingMessage):
        raise NotImplementedError



from typing import Literal, TypeAlias

from pydantic import BaseModel, Field

from aioka.server.schema import BaseMeta

ErrorTypes: TypeAlias = Literal["validation", "internal"]


class ExceptionSchema(BaseModel):
    response_ok: bool = Field(description="True если запрос обработан успешно")
    error_type: ErrorTypes = Field(description="Тип ошибки")
    validation_details: dict = Field(
        {}, description="Детали для ошибки валидации"
    )
    internal_details: dict = Field({}, description="Детали для других ошибок")
    raw_message: str | None = Field(
        None,
        description="Сообщение, которое не удалось обработать (для rmq)",
    )


class FullErrorSchema(BaseModel):
    meta: BaseMeta
    payload: ExceptionSchema



from functools import wraps
from typing import AsyncIterable, Awaitable, Callable, Type

from pydantic import BaseModel

from aioka.action.base import BaseAction
from aioka.action.master import Action
from aioka.server import RequestMeta, ResponseMeta
from aioka.server.base import BaseServer


class BaseHandler:
    """
    Basic implementation of the handler.
    Binds actions to rft_proxy calls and initializes the client
    """

    __slots__ = ("routes", "server")

    server: BaseServer

    def __init__(self):
        self.routes: dict[str, Type[BaseAction]] = {}

    @classmethod
    async def create(cls, server: BaseServer):
        """
        Фабрика для асинхронного создания инстанса класса

        :param server: Кастомный сервер, базирующейся на BaseServer
        """
        self = BaseHandler()
        self.server = server
        return self

    async def _bind_procedures(self) -> AsyncIterable:
        for route, action in self.routes.items():
            action_instance: BaseAction = action(action_name=route)
            action_instance.service_name = self.server.service_name
            yield action_instance

    async def bind(self):
        """Метод для биндинга процедур к серверу"""
        async for procedure in self._bind_procedures():
            await self.server.bind(procedure.action_name, procedure)

    def add_action(self, action_name: str, action: BaseAction):
        self.routes.update({action_name: action})

    def action(
        self,
        action_name: str,
        request_payload: BaseModel = BaseModel,
        response_payload: BaseModel = BaseModel,
        request_meta: RequestMeta = RequestMeta,
        response_meta: ResponseMeta = ResponseMeta,
    ):
        def decorator(
            func: Callable[
                [BaseModel, BaseModel],
                Awaitable[BaseModel | dict | list | None],
            ]
        ) -> BaseAction:
            @wraps(func)
            async def wrapper(meta: request_meta, payload: request_payload):
                return await func(meta, payload)

            action = Action(action_name=action_name)
            action.__call__ = wrapper
            action.REQUEST_PAYLOAD = request_payload
            action.RESPONSE_PAYLOAD = response_payload
            action.REQUEST_META = request_meta
            action.RESPONSE_META = response_meta

            self.add_action(action_name=action_name, action=action)
            return action

        return decorator



import asyncio
import logging
from typing import Optional

import aio_pika
from aio_pika import Channel, DeliveryMode, Message
from aio_pika.abc import (
    AbstractExchange,
    AbstractQueue,
    AbstractRobustConnection,
    ExchangeType,
)
from aio_pika.pool import Pool
from aiormq.tools import awaitable

from aioka.action.base import BaseAction
from aioka.server.base import BaseServer

logger = logging.getLogger(__name__)


class Worker:
    __slots__ = (
        "queue",
        "consumer_tag",
        "loop",
    )

    def __init__(self, queue: AbstractQueue, consumer_tag: str, loop):
        self.queue = queue
        self.consumer_tag = consumer_tag
        self.loop = loop

    def close(self) -> asyncio.Task:
        """Cancel subscription to the channel"""

        async def closer():
            await self.queue.cancel(self.consumer_tag)

        return self.loop.create_task(closer())


class Master:
    """
    Change serialize/deserialize methods for json messages
    """

    __slots__ = (
        "channel",
        "_loop",
        "proxy",
        "_workers",
        "default_exchange",
        "service_name",
        "queue_error_name",
        "exchange",
        "exchange_type",
    )

    def __init__(
        self,
        service_name: str,
        queue_error_name: str,
        channel: Channel,
        exchange: str,
        exchange_type: ExchangeType,
    ):
        self.service_name = service_name
        self.queue_error_name = queue_error_name
        self.channel = channel
        self._workers: list[Worker] = []
        self._loop = asyncio.get_event_loop()
        self.exchange = exchange
        self.exchange_type = exchange_type

    async def create_queue(self, channel_name, **kwargs) -> AbstractQueue:
        return await self.channel.declare_queue(channel_name, **kwargs)

    async def create_worker(
        self, channel_name: str, procedure: BaseAction, **kwargs
    ) -> Worker:
        queue = await self.create_queue(channel_name, **kwargs)

        await procedure.set_client(self.publish)
        if hasattr(procedure.consume, "_is_coroutine"):
            fn = procedure.consume
        else:
            fn = awaitable(procedure.consume)

        consumer_tag = await queue.consume(fn)
        worker = Worker(queue, consumer_tag, self._loop)
        await worker.queue.bind(
            self.default_exchange, routing_key=channel_name
        )
        self._workers.append(worker)
        return worker

    async def publish(
        self,
        message: bytes,
        routing_key: Optional[str] = None,
        delivery_mode: DeliveryMode = DeliveryMode.PERSISTENT,
        content_type: str = "application/json",
        mandatory: bool = False,  # Check routing_key exists
        immediate: bool = False,  # Request immediate delivery
        exchange_name: str = None,
        **message_kwargs,
    ):
        if not exchange_name:
            exchange: AbstractExchange = self.default_exchange
        else:
            exchange: AbstractExchange = await self.channel.get_exchange(
                exchange_name
            )

        message = Message(
            body=message,
            content_type=content_type,
            delivery_mode=delivery_mode,
            **message_kwargs,
        )

        if routing_key is None:
            routing_key = f"{self.service_name}.{self.queue_error_name}"
        await exchange.publish(
            message=message,
            routing_key=routing_key,
            mandatory=mandatory,
            immediate=immediate,
        )

    async def __aenter__(self):
        self.default_exchange = await self.channel.declare_exchange(
            self.exchange,
            self.exchange_type,
            durable=True,
        )
        error_queue_name = f"{self.service_name}" f".{self.queue_error_name}"
        error_queue = await self.create_queue(error_queue_name)
        await error_queue.bind(
            self.default_exchange, routing_key=error_queue_name
        )
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        for worker in self._workers:
            worker.close()

    @property
    def workers(self):
        return self._workers


class AsyncRqMasterServer(BaseServer):
    """
    Asynchronous server for working with queues
     using the Master-Worker pattern
    """

    def __init__(
        self,
        rmq_urls: list[str],
        service_name: str,
        connection_pool_size: int = 2,
        channel_pool_size: int = 10,
    ):
        super().__init__(rmq_urls, service_name)
        self.default_exchange: AbstractExchange | None = None
        self.master: Master | None = None
        self.connection_pool_size = connection_pool_size
        self.channel_pool_size = channel_pool_size

    @classmethod
    async def create(
        cls,
        exchange: str = "",
        exchange_type: ExchangeType = ExchangeType.DIRECT,
        prefetch_count: int = 0,
        prefetch_size: int = 0,
        durable: bool = True,
        **kwargs,
    ):
        """Fabric for async init class"""
        self = AsyncRqMasterServer(**kwargs)

        async def get_connection() -> AbstractRobustConnection:
            return await aio_pika.connect_robust(self.rmq_urls[0])

        async def get_channel() -> aio_pika.Channel:
            async with self.connection_pool.acquire() as connection:
                return await connection.channel()

        self.connection_pool = Pool(
            get_connection, max_size=self.connection_pool_size
        )
        self.channel_pool = Pool(get_channel, max_size=self.channel_pool_size)

        async with self.channel_pool.acquire() as channel:  # type: Channel
            await channel.set_qos(
                prefetch_count=prefetch_count, prefetch_size=prefetch_size
            )
            async with Master(
                channel=channel,
                service_name=self.service_name,
                queue_error_name=self.queue_error_name,
                exchange=exchange,
                exchange_type=exchange_type,
            ) as master:  # type: Master
                self.master = master
                self.default_exchange = (
                    await self.master.channel.declare_exchange(
                        name=exchange,
                        type=exchange_type,
                        durable=durable,
                    )
                )
                error_queue = await self.master.create_queue(
                    f"{self.service_name}.{self.queue_error_name}"
                )
                await error_queue.bind(self.default_exchange)
                return self

    async def bind(self, method_name: str, procedure: BaseAction):
        """Method for creating queue and binding actions"""
        worker = await self.master.create_worker(
            f"{self.service_name}.{method_name}", procedure
        )
        await worker.queue.bind(self.default_exchange)

    async def shutdown(self):
        for worker in self.master.workers:
            worker.close()
